import os
from .._models import ProxyMap as _Proxy
from ..pac import load_pac as _load_pac
from ..utils import SingleProxyMap as _Simple
from .common import EnvProxyConfig

if os.name == "nt":
    from .windows import system_proxy
else:
    def system_proxy() -> "_Proxy|str":
        return EnvProxyConfig.from_env()  


def auto_proxy(**urlopen_kwargs) -> _Proxy:
    proxy = system_proxy()
    if isinstance(proxy, str):
        return _load_pac(proxy, **urlopen_kwargs)
    else:
        return proxy



from typing import TYPE_CHECKING, Sequence, TypedDict as _Dict
from socket import AddressFamily as _A
from ._models import ProxyMap, Proxy
from .pac import load_pac

class SingleProxyMap(ProxyMap):
    def __init__(self, proxy: 'Proxy|Sequence[Proxy]' = None) -> None:
        self.proxies:Sequence[Proxy] = proxy if isinstance(proxy, Sequence) else (proxy,)

    def __getitem__(self):
        return self.proxies


try:
    import netifaces

    if TYPE_CHECKING:
        from netifaces import INETAddress

    def get_addresses():
        ints: "list[INETAddress]" = []
        for i in netifaces.interfaces():
            addrs = netifaces.interfaces(i)
            ints.extend(addrs.get(_A.AF_INET, []))
        return ints

except ImportError:

    def get_addresses():
        return []

def get_proxymap_for(src: "str|Proxy|None") -> ProxyMap:
    if isinstance(src, str):
        proxy = Proxy.from_uri(src)
        netloc = proxy.netloc

        if (
            proxy.scheme in ["http", "https"]
            and not src.endswith(netloc)
            or src.endswith(netloc + "/")
        ):
            return load_pac(src)
    else:
        proxy = src

    return SingleProxyMap(proxy)



from typing import (
    Iterable as _Iter,
    Literal as _Literal,
    get_args as _args,
    cast as _cast,
    overload as _overload,
)
import socket as _socket
import ipaddress as _ip
from fnmatch import fnmatch as _shexpmatch
import datetime as _datetime
from urllib.request import urlopen as _urlopen
from warnings import warn as _warn
from ._models import ProxyMap, Proxy
from ._re import _PAC_REGEX, _URI_REGEX


_WEEKDAY = _Literal["SUN", "MON", "TUE", "WED", "THU", "FRI", "SAT"]
_WEEKDAYS = ("SUN", "MON", "TUE", "WED", "THU", "FRI", "SAT")


class PAC(ProxyMap):

    #### UTILITY FUNCTIONS ####
    @staticmethod
    def dnsResolve(host: str, /):
        try:
            return _socket.gethostbyname(host)
        except:
            return

    @staticmethod
    def myIpAddress():
        return _socket.gethostbyname(_socket.gethostname())

    @staticmethod
    def dnsDomainLevels(host: str, /):
        return len(host.split("."))

    @staticmethod
    def convert_addr(ipaddr: str, /):
        return int(_ip.ip_address(ipaddr))

    @staticmethod
    def shExpMatch(test: str, shexp: str, /):
        return _shexpmatch(test, shexp)

    #### TIME FUNCTIONS ####
    @_overload
    def weekdayRange(wd1: _WEEKDAY, gmt: 'None|_Literal["GMT"]' = None, /):
        ...

    @_overload
    def weekdayRange(
        wd1: _WEEKDAY, wd2: _WEEKDAY, gmt: 'None|_Literal["GMT"]' = None, /
    ):
        ...

    @staticmethod
    def weekdayRange(wd1: _WEEKDAY, /, *args: '_WEEKDAY|_Literal["GMT"]'):
        start = _WEEKDAYS.index(wd1.upper())
        if args:
            wd2 = args[0].upper()
            if len(args) == 2:
                gmt = args[1].upper() == "GMT"
                end = _WEEKDAYS[wd2]
            elif wd2 == "GMT":
                gmt = True
                end = start
            else:
                end = _WEEKDAYS[wd2]

        else:
            end = start

        today = (
            _datetime.datetime.utcnow() if gmt else _datetime.datetime.now()
        ).isoweekday()
        if today == 7:
            today = 0
        return start <= today <= end

    @staticmethod
    def dateRange(*args):
        return False

    @staticmethod
    def timeRange(*args):
        return False

    #### HOSTNAME FUNCTIONS ####

    @staticmethod
    def isPlainHostname(host: str):
        return "." not in host

    @staticmethod
    def dnsDomainIs(host: str, domain: str):
        return host.endswith(domain)

    @staticmethod
    def localHostOrDomainIs(host: str, hostdom: str):
        return "." not in host and hostdom.startswith(host) or hostdom == host

    @staticmethod
    def isResolvable(host: str):
        try:
            _socket.gethostbyname(host)
            return True
        except:
            return False

    @staticmethod
    def isInNet(host: str, pattern: str, mask: str):
        try:
            ip = _ip.IPv4Address(host)
        except:
            try:
                ip = _ip.IPv4Address(PAC.dnsResolve(host))
            except:
                return False
        net = _ip.IPv4Network(f"{pattern}/{mask}", strict=False)
        return ip in net

    @staticmethod
    def FindProxyForURL(url: str, host: str, /) -> str:
        return "DIRECT"

    def __getitem__(self, url: str) -> _Iter[Proxy]:
        parsed = _URI_REGEX.match(url)
        if not parsed:
            raise ValueError(url)
        scheme, user, p, host, port = parsed.groups()
        pac_proxies = self.FindProxyForURL(
            f"{scheme}://{host}{f':{port}' if port else ''}", host
        )
        return (Proxy(*proxy) for proxy in _PAC_REGEX.findall(pac_proxies))


try:
    from ._jscontext import JSContext as _JSContext

    class JSProxyAutoConfig(PAC, _JSContext):
        ...

    _jspac = True
except ImportError:

    _jspac = False


def load_pac(url: str, **urllib_kwds):
    if "://" not in url:
        url = ("file://" if url.startswith("/") else "https://") + url
    with _urlopen(url, **urllib_kwds) as resp:
        js = resp.read()
    if "FindProxyForURL" not in js:
        raise Exception("Not FindProxyForURL found int response from: " + url)
    if not _jspac:
        _warn(f"Can not load js from: {url} as pac. Install libproxy[jspac]")
        return PAC()
    else:
        return JSProxyAutoConfig(js)



from abc import ABC
from socket import getservbyname
from typing import Iterable, NamedTuple, overload

from ._re import _URI_REGEX


class _URI(NamedTuple):
    scheme: str
    username: str
    password: str
    host: str
    port: "int|None"

    @property
    def netloc(self):
        if self.port:
            return f"{self.host}:{self.port}"
        else:
            return self.host

    def resolved(self):
        if self.port:
            return self
        else:
            self.__class__(
                self.scheme,
                self.username,
                self.password,
                self.host,
                getservbyname(self.scheme),
            )

    @classmethod
    def from_str(cls, uri: str):
        return cls(*_URI_REGEX.match(uri).groups()) if uri else None


class URL(_URI):
    _DEFAULT_SCHEME = "http"

    def __new__(
        cls, scheme: str, username: str, password: str, host: str, port: str
    ) -> "Proxy":
        scheme = scheme.lower()
        if not scheme:
            scheme = cls._DEFAULT_SCHEME

        if port:
            port = int(port)

        return super().__new__(cls, scheme, username, password, host, port)


class Proxy(_URI):
    _DEFAULT_SCHEME = "http"

    def __new__(
        cls, scheme: str, username: str, password: str, host: str, port: str
    ) -> "Proxy":
        scheme = scheme.lower()
        if scheme == "direct":
            return None
        elif scheme == "proxy":
            scheme = "http"
        elif scheme == "socks":
            scheme = "socks4"
        elif not scheme:
            scheme = cls._DEFAULT_SCHEME

        if port:
            port = int(port)

        return super().__new__(cls, scheme, username, password, host, port)

    @staticmethod
    def from_uris(uri: str):
        return set([Proxy(*proxy) for proxy in _URI_REGEX.findall(uri)] if uri else [])

    @property
    def url(self):
        return f"{self.scheme}://{self.netloc}"

class ProxyMap(ABC):
    @overload
    def __getitem__(self, uri: str) -> Iterable[Proxy]:
        pass



import winreg as _reg
from .common import  EnvProxyConfig
from ..utils import SingleProxyMap
def _queryvalue(key:_reg.HKEYType, subkey:str)->str:
    try:
        value, type = _reg.QueryValueEx(key, subkey)
        return value
    except:
        return None

_WINREG_INTERNET_SETTINGS = r"SOFTWARE\Microsoft\Windows\CurrentVersion\Internet Settings"

def system_proxy():
    settings = _reg.OpenKey(_reg.HKEY_CURRENT_USER, _WINREG_INTERNET_SETTINGS)
    pac = _queryvalue(settings, "AutoConfigURL")
    if pac:
        return pac
    if _queryvalue(settings, "ProxyEnable"):
        env_proxy = _queryvalue(settings, "ProxyServer")
        overrides = _queryvalue(settings, "ProxyOverride")
        if env_proxy:
            return EnvProxyConfig(env_proxy, env_proxy, overrides.split(";") if overrides else [] )
    
    if _queryvalue(settings, "EnableNegotiate"):
        #TODO
        pass
    return SingleProxyMap()




import os
from typing import Iterable as _Iter
import re
from .._models import Proxy, ProxyMap, URL
from ..pac import load_pac, PAC
from ..utils import SingleProxyMap, get_proxymap_for, get_addresses


class EnvProxyConfig(ProxyMap):
    __slots__ = ("http_proxy", "https_proxy", "no_proxy")

    def __init__(
        self,
        http_proxy: "str|Proxy|None",
        https_proxy: "str|Proxy|None",
        no_proxy: "_Iter[str]",
    ) -> None:
        self.http_proxy = get_proxymap_for(http_proxy)
        self.https_proxy = get_proxymap_for(https_proxy)
        self.no_proxy = (
            [
                re.compile(re.escape(_no) + ".*") if _no != "<local>" else None
                for _no in set(no_proxy)
            ]
            if no_proxy
            else []
        )

    def __getitem__(self, url: str) -> _Iter[Proxy]:
        uri = URL.from_str(url)
        url = f"{uri.scheme}://{uri.netloc}"
        for _no in self.no_proxy:
            if _no is None:
                for addr in get_addresses():
                    if PAC.isInNet(url, addr["addr"], addr["netmask"]):
                        return [None]
            else:
                if _no.match(url):
                    return [None]
        return self.https_proxy[url] if uri.scheme == "https" else self.http_proxy[url]

    @staticmethod
    def from_env():
        https = os.environ.get("HTTPS_PROXY", None)
        if not https:
            https = os.environ.get("https_proxy")

        http = os.environ.get("HTTP_PROXY", None)
        if not http:
            http = os.environ.get("http_proxy")

        no_proxy = os.environ.get("NO_PROXY", None)
        if not no_proxy:
            no_proxy = os.environ.get("no_proxy")

        return EnvProxyConfig(http, https, no_proxy)



ALPHA = r"A-Za-z"
DIGIT = r"0-9"
import re
SCHEME = rf"[{ALPHA}][{ALPHA}{DIGIT}+-.]*"
PORT = rf"[{DIGIT}]*"
NON_BREAKING = rf"[^:@/;]"
AUTHORITY = (
    rf"(?:({NON_BREAKING}*)(?::({NON_BREAKING}*))?@)?({NON_BREAKING}+)(?::({PORT}))?"
)
DELIM = r"(?:;|^)\s*"
_PAC_REGEX = re.compile(rf"{DELIM}({SCHEME})(?:\s+(?:{AUTHORITY})?\s*)")
_URI_REGEX = re.compile(rf"{DELIM}(?:(?:({SCHEME})://)?(?:{AUTHORITY})?\s*)")


from ast import arg
from pathlib import Path
import shutil, sys
from setuptools import setup as _setup, find_packages


root = Path(__file__).parent
PKG = sys.argv.pop(1).replace("-", "_")
PKG_NAME = PKG.replace("_", "-")

src = root / "src" / PKG
dist = root / "dist" / PKG
dist.mkdir(exist_ok=True, parents=True)


def clean_dist():
    if PKG == "_":
        for path in (root / "dist").iterdir():
            shutil.rmtree(path)
    else:
        for path in dist.iterdir():
            path.unlink()


def clean():

    for path in (root / "src").iterdir():
        if path.suffix == ".egg-info":
            shutil.rmtree(path)
    for path in root.iterdir():
        if path.suffix == ".egg-info":
            shutil.rmtree(path)
    shutil.rmtree(root / "build", ignore_errors=True)


if sys.argv[1] == "clean-dist":
    clean_dist()
    exit()


pkgs = find_packages(str(src))
readme_file = next((f for f in src.iterdir() if f.stem == "README"), None)
if readme_file:
    readme = readme_file.read_text()

    if readme_file.suffix == ".rst":
        readme_type = f"text/x-rst"
    elif readme_file.suffix == ".md":
        readme_type = "text/markdown"
    else:
        readme_type = "text/plain"
else:
    readme = None
    readme_type = None
print(readme_type)


def setup(**kwargs):
    _setup(
        name=PKG.replace("_", "-"),
        version=(src / "VERSION").read_text(),
        description=(src / "DESCRIPTION").read_text(),
        long_description=readme,
        long_description_content_type=readme_type
        if readme_type != "text/x-rst"
        else None,
        author="Jose A.",
        author_email="jose-pr@coqui.dev",
        url=f"https://github.com/jose-pr/{root.resolve().name}",
        package_dir={"": "src"},
        packages=[PKG, *[f"{PKG}.{pkg}" for pkg in pkgs]],
        install_requires=(src / "requirements.txt").read_text().splitlines(),
        **kwargs,
    )


clean()
if sys.argv[1] == "dist-build":
    sys.argv.append("--dist-dir")
    sys.argv.append(f"dist/{PKG}")
    clean_dist()
    sys.argv[1] = "bdist_wheel"
    setup()
    sys.argv[1] = "sdist"
    setup()
else:
    setup()

clean()



from abc import ABCMeta
from typing import OrderedDict, Sequence
from dukpy import JSInterpreter

class JSContextMeta(ABCMeta):
    def __new__(
        metaclass: "type[JSContext]",
        cls_name: str,
        base_classes: Sequence[object],
        cls_builder: OrderedDict[str, object],
    ):
        jsContext: "dict[str, object]" = cls_builder.pop("_JSCONTEXT", {})
        exclude: "list[str]" = cls_builder.get("_JSCONTEXT_EXCLUDE", [])
        for key, val in cls_builder.items():
            if key[0].isalpha():
                jsContext.setdefault(key, val)

        for cls in reversed(base_classes):
            exclude.extend(getattr(cls, "_JSCONTEXT_EXCLUDE", []))
            if hasattr(cls, "_JSCONTEXT"):
                update: dict = cls._JSCONTEXT
            else:
                update = {key: getattr(cls, key) for key in dir(cls) if key[0].isalpha()}
            for key, val in update.items():
                jsContext.setdefault(key, staticmethod(val))

        return type.__new__(
            metaclass,
            cls_name,
            base_classes,
            {
                **{
                    key: val
                    for key, val in cls_builder.items()
                    if not (key[0].isalpha() and key[0] not in exclude)
                },
                "_JSCONTEXT": {
                    key: val
                    for key, val in jsContext.items()
                    if (key[0].isalpha() and key[0] not in exclude)
                },
            },
        )


class JSContext(metaclass=JSContextMeta):
    def __init__(self, js:str) -> None:
        context:dict = object.__getattribute__(self, "_JSCONTEXT")
        engine = JSInterpreter()
        for key, val in context.items():
            if isinstance(val, staticmethod):
                val = val.__func__
            elif isinstance(val, classmethod):
                val = val.__get__(self.__class__)
            elif isinstance(val, property):
                raise Exception("Not supported yet")
            else:
                val = val.__get__(self)

            engine.export_function(key, val)
            _js = f"function {key}(){{ return call_python.apply(null, ['{key}'].concat(Array.prototype.slice.call(arguments))); }}"
            engine.evaljs(_js)
        engine.evaljs(js)

        object.__setattr__(self, "_jsengine", engine)

    def __getattribute__(self, name: str):
        engine:JSInterpreter = object.__getattribute__(self, "_jsengine")
        context:dict = object.__getattribute__(self, "_JSCONTEXT")
        if name in context:
            def jsFunction(*args):
                return engine.evaljs(f"{name}.apply(null, dukpy.args)", args = list(args))
            return jsFunction
        else:
            return object.__getattribute__(self, name)
